@@ -18,6 +18,7 @@
 #include "lib/device/device.h"
 #include "lib/display/display.h"
 #include "lib/label/label.h"
+#include "lib/device/filesystem.h"
 
 #define NUMBER_OF_MAJORS 4096
 
@@ -37,15 +38,15 @@ struct dev_type_def {
 };
 
 struct dev_types {
-	int md_major;
-	int blkext_major;
-	int drbd_major;
-	int device_mapper_major;
-	int emcpower_major;
-	int vxdmp_major;
-	int power2_major;
-	int dasd_major;
-	int loop_major;
+	unsigned md_major;
+	unsigned blkext_major;
+	unsigned drbd_major;
+	unsigned device_mapper_major;
+	unsigned emcpower_major;
+	unsigned vxdmp_major;
+	unsigned power2_major;
+	unsigned dasd_major;
+	unsigned loop_major;
 	struct dev_type_def dev_type_array[NUMBER_OF_MAJORS];
 };
 
@@ -99,9 +100,11 @@ int dev_is_pmem(struct dev_types *dt, struct device *dev);
 
 int dev_is_nvme(struct dev_types *dt, struct device *dev);
 
-int dev_is_lv(struct device *dev);
+int dev_is_lv(struct cmd_context *cmd, struct device *dev);
 
-int get_fs_block_size(const char *pathname, uint32_t *fs_block_size);
+#define FSTYPE_MAX 16
+int fs_block_size_and_type(const char *pathname, uint32_t *fs_block_size_bytes, char *fstype, int *nofs);
+int fs_get_blkid(const char *pathname, struct fs_info *fsi);
 
 int dev_is_used_by_active_lv(struct cmd_context *cmd, struct device *dev, int *used_by_lv_count,
 			     char **used_by_dm_name, char **used_by_vg_uuid, char **used_by_lv_uuid);
