@@ -32,7 +32,7 @@ static const struct {
 	{
 	ALLOC_CONTIGUOUS, "contiguous", 'c'}, {
 	ALLOC_CLING, "cling", 'l'}, {
-	ALLOC_CLING_BY_TAGS, "cling_by_tags", 't'}, {	/* Only used in log mesgs */
+	ALLOC_CLING_BY_TAGS, "cling_by_tags", 't'}, {	/* Only used in log messages */
 	ALLOC_NORMAL, "normal", 'n'}, {
 	ALLOC_ANYWHERE, "anywhere", 'a'}, {
 	ALLOC_INHERIT, "inherit", 'i'}
@@ -95,6 +95,8 @@ const char *get_lock_type_string(lock_type_t lock_type)
 		return "dlm";
 	case LOCK_TYPE_SANLOCK:
 		return "sanlock";
+	case LOCK_TYPE_IDM:
+		return "idm";
 	}
 	return "invalid";
 }
@@ -111,13 +113,15 @@ lock_type_t get_lock_type_from_string(const char *str)
 		return LOCK_TYPE_DLM;
 	if (!strcmp(str, "sanlock"))
 		return LOCK_TYPE_SANLOCK;
+	if (!strcmp(str, "idm"))
+		return LOCK_TYPE_IDM;
 	return LOCK_TYPE_INVALID;
 }
 
-static const char *_percent_types[7] = { "NONE", "VG", "FREE", "LV", "PVS", "ORIGIN" };
-
 const char *get_percent_string(percent_type_t def)
 {
+	static const char _percent_types[][8] = { "NONE", "VG", "FREE", "LV", "PVS", "ORIGIN" };
+
 	return _percent_types[def];
 }
 
@@ -379,7 +383,7 @@ static int _lvdisplay_historical_full(struct cmd_context *cmd,
 	log_print("--- Historical Logical volume ---");
 
 	if (lvm1compat)
-		/* /dev/vgname/lvname doen't actually exist for historical devices */
+		/* /dev/vgname/lvname doesn't actually exist for historical devices */
 		log_print("LV Name                %s%s/%s",
 			  hlv->vg->cmd->dev_dir, hlv->vg->name, hlv->name);
 	else
@@ -399,7 +403,7 @@ int lvdisplay_full(struct cmd_context *cmd,
 		   void *handle __attribute__((unused)))
 {
 	struct lvinfo info;
-	int inkernel, snap_active = 0;
+	int inkernel, snap_active = 0, partial = 0, raid_is_avail = 1;
 	char uuid[64] __attribute__((aligned(8)));
 	const char *access_str;
 	struct lv_segment *snap_seg = NULL, *mirror_seg = NULL;
@@ -435,7 +439,7 @@ int lvdisplay_full(struct cmd_context *cmd,
 	lvm1compat = find_config_tree_bool(cmd, global_lvdisplay_shows_full_device_path_CFG, NULL);
 
 	if (lvm1compat)
-		/* /dev/vgname/lvname doen't actually exist for internal devices */
+		/* /dev/vgname/lvname doesn't actually exist for internal devices */
 		log_print("LV Name                %s%s/%s",
 			  lv->vg->cmd->dev_dir, lv->vg->name, lv->name);
 	else if (lv_is_visible(lv)) {
@@ -475,7 +479,7 @@ int lvdisplay_full(struct cmd_context *cmd,
 					  snap_active ? "active" : "INACTIVE");
 		}
 		snap_seg = NULL;
-	} else if ((snap_seg = find_snapshot(lv))) {
+	} else if (lv_is_cow(lv) && (snap_seg = find_snapshot(lv))) {
 		if (inkernel &&
 		    (snap_active = lv_snapshot_percent(snap_seg->cow,
 						       &snap_percent)))
@@ -558,11 +562,18 @@ int lvdisplay_full(struct cmd_context *cmd,
 		log_print("LV VDO Pool name       %s", seg_lv(seg, 0)->name);
 	}
 
+	if (lv_is_partial(lv))
+		partial = 1;
+
+	if (lv_is_raid(lv))
+		raid_is_avail = raid_is_available(lv) ? 1 : 0;
+
 	if (inkernel && info.suspended)
 		log_print("LV Status              suspended");
 	else if (activation())
-		log_print("LV Status              %savailable",
-			  inkernel ? "" : "NOT ");
+		log_print("LV Status              %savailable%s",
+			  (inkernel && raid_is_avail) ? "" : "NOT ",
+			  partial ? " (partial)" : "");
 
 /********* FIXME lv_number
     log_print("LV #                   %u", lv->lv_number + 1);
@@ -925,7 +936,8 @@ char yes_no_prompt(const char *prompt, ...)
 	static const char _no[] = "no";
 	const char *answer = NULL;
 	int c = silent_mode() ? EOF : 0;
-	int i = 0, ret = 0, sig = 0;
+	int ret = 0, sig = 0;
+	unsigned i = 0;
 	char buf[12];
 	va_list ap;
 
@@ -963,7 +975,7 @@ char yes_no_prompt(const char *prompt, ...)
 
 		c = tolower(c);
 
-		if ((ret > 0) && (c == answer[0]))
+		if ((ret > 0) && answer && (c == answer[0]))
 			answer++;	/* Matching, next char */
 		else if (c == '\n') {
 			if (feof(stdin))
@@ -987,9 +999,10 @@ char yes_no_prompt(const char *prompt, ...)
 			answer = _no + 1;	/* Expecting 'No' */
 		} else if (!ret && isspace(c)) {
 			/* Ignore any whitespace before */
-			--i;
+			if (i > 0)
+				i--;
 			goto nextchar;
-		} else if ((ret > 0) && isspace(c)) {
+		} else if ((ret > 0) && answer && isspace(c)) {
 			/* Ignore any whitespace after */
 			while (*answer)
 				answer++; /* jump to end-of-word */
