@@ -128,7 +128,7 @@ static int _read_id(struct id *id, const struct dm_config_node *cn, const char *
 	return 1;
 }
 
-static int _read_flag_config(const struct dm_config_node *n, uint64_t *status, int type)
+static int _read_flag_config(const struct dm_config_node *n, uint64_t *status, enum pv_vg_lv_e type)
 {
 	const struct dm_config_value *cv;
 	*status = 0;
@@ -159,7 +159,7 @@ static int _read_str_list(struct dm_pool *mem, struct dm_list *list, const struc
 	if (cv->type == DM_CFG_EMPTY_ARRAY)
 		return 1;
 
-	while (cv) {
+	do {
 		if (cv->type != DM_CFG_STRING) {
 			log_error("Found an item that is not a string");
 			return 0;
@@ -168,8 +168,7 @@ static int _read_str_list(struct dm_pool *mem, struct dm_list *list, const struc
 		if (!str_list_add(mem, list, dm_pool_strdup(mem, cv->v.str)))
 			return_0;
 
-		cv = cv->next;
-	}
+	} while ((cv = cv->next));
 
 	return 1;
 }
@@ -188,7 +187,7 @@ static int _read_pv(struct cmd_context *cmd,
 	struct physical_volume *pv;
 	struct pv_list *pvl;
 	const struct dm_config_value *cv;
-	const char *device_hint;
+	const char *str;
 	uint64_t size, ba_start;
 
 	if (!(pvl = dm_pool_zalloc(mem, sizeof(*pvl))) ||
@@ -219,7 +218,8 @@ static int _read_pv(struct cmd_context *cmd,
 	if (!(pv->vg_name = dm_pool_strdup(mem, vg->name)))
 		return_0;
 
-	memcpy(&pv->vgid, &vg->id, sizeof(vg->id));
+	/* both are struct id */
+	memcpy(&pv->vg_id, &vg->id, sizeof(struct id));
 
 	if (!_read_flag_config(pvn, &pv->status, PV_FLAGS)) {
 		log_error("Couldn't read status flags for physical volume.");
@@ -233,11 +233,21 @@ static int _read_pv(struct cmd_context *cmd,
 		return 0;
 	}
 
-	if (dm_config_get_str(pvn, "device", &device_hint)) {
-		if (!(pv->device_hint = dm_pool_strdup(mem, device_hint)))
+	if (dm_config_get_str(pvn, "device", &str)) {
+		if (!(pv->device_hint = dm_pool_strdup(mem, str)))
 			log_error("Failed to allocate memory for device hint in read_pv.");
 	}
 
+	if (dm_config_get_str(pvn, "device_id", &str)) {
+		if (!(pv->device_id = dm_pool_strdup(mem, str)))
+			log_error("Failed to allocate memory for device_id in read_pv.");
+	}
+
+	if (dm_config_get_str(pvn, "device_id_type", &str)) {
+		if (!(pv->device_id_type = dm_pool_strdup(mem, str)))
+			log_error("Failed to allocate memory for device_id_type in read_pv.");
+	}
+
 	if (!_read_uint64(pvn, "pe_start", &pv->pe_start)) {
 		log_error("Couldn't read extent start value (pe_start) "
 			  "for physical volume.");
@@ -306,7 +316,7 @@ static int _read_pvsummary(struct cmd_context *cmd,
 {
 	struct physical_volume *pv;
 	struct pv_list *pvl;
-	const char *device_hint;
+	const char *str;
 
 	if (!(pvl = dm_pool_zalloc(mem, sizeof(*pvl))) ||
 	    !(pvl->pv = dm_pool_zalloc(mem, sizeof(*pvl->pv))))
@@ -326,9 +336,19 @@ static int _read_pvsummary(struct cmd_context *cmd,
 	    !_read_uint64(pvn, "dev_size", &pv->size))
 		log_warn("Couldn't read dev size for physical volume.");
 
-	if (dm_config_get_str(pvn, "device", &device_hint)) {
-		if (!(pv->device_hint = dm_pool_strdup(mem, device_hint)))
-			log_error("Failed to allocate memory for device hint in read_pv.");
+	if (dm_config_get_str(pvn, "device", &str)) {
+		if (!(pv->device_hint = dm_pool_strdup(mem, str)))
+			log_error("Failed to allocate memory for device hint in read_pv_sum.");
+	}
+
+	if (dm_config_get_str(pvn, "device_id", &str)) {
+		if (!(pv->device_id = dm_pool_strdup(mem, str)))
+			log_error("Failed to allocate memory for device_id in read_pv_sum.");
+	}
+
+	if (dm_config_get_str(pvn, "device_id_type", &str)) {
+		if (!(pv->device_id_type = dm_pool_strdup(mem, str)))
+			log_error("Failed to allocate memory for device_id_type in read_pv_sum.");
 	}
 
 	dm_list_add(&vgsummary->pvsummaries, &pvl->list);
@@ -356,7 +376,8 @@ static int _read_segment(struct cmd_context *cmd,
 			 struct format_instance *fid,
 			 struct dm_pool *mem,
 			 struct logical_volume *lv, const struct dm_config_node *sn,
-			 struct dm_hash_table *pv_hash)
+			 struct dm_hash_table *pv_hash,
+			 struct dm_hash_table *lv_hash)
 {
 	uint32_t area_count = 0u;
 	struct lv_segment *seg;
@@ -429,7 +450,7 @@ static int _read_segment(struct cmd_context *cmd,
 	}
 
 	if (seg->segtype->ops->text_import &&
-	    !seg->segtype->ops->text_import(seg, sn_child, pv_hash))
+	    !seg->segtype->ops->text_import(seg, sn_child, pv_hash, lv_hash))
 		return_0;
 
 	/* Optional tags */
@@ -531,7 +552,8 @@ static int _read_segments(struct cmd_context *cmd,
 			  struct format_instance *fid,
 			  struct dm_pool *mem,
 			  struct logical_volume *lv, const struct dm_config_node *lvn,
-			  struct dm_hash_table *pv_hash)
+			  struct dm_hash_table *pv_hash,
+			  struct dm_hash_table *lv_hash)
 {
 	const struct dm_config_node *sn;
 	int count = 0, seg_count;
@@ -542,7 +564,7 @@ static int _read_segments(struct cmd_context *cmd,
 		 * All sub-sections are assumed to be segments.
 		 */
 		if (!sn->v) {
-			if (!_read_segment(cmd, fmt, fid, mem, lv, sn, pv_hash))
+			if (!_read_segment(cmd, fmt, fid, mem, lv, sn, pv_hash, lv_hash))
 				return_0;
 
 			count++;
@@ -959,7 +981,7 @@ static int _read_lvsegs(struct cmd_context *cmd,
 
 	memcpy(&lv->lvid.id[0], &lv->vg->id, sizeof(lv->lvid.id[0]));
 
-	if (!_read_segments(cmd, fmt, fid, mem, lv, lvn, pv_hash))
+	if (!_read_segments(cmd, fmt, fid, mem, lv, lvn, pv_hash, lv_hash))
 		return_0;
 
 	lv->size = (uint64_t) lv->le_count * (uint64_t) vg->extent_size;
@@ -1051,19 +1073,19 @@ static struct volume_group *_read_vg(struct cmd_context *cmd,
 	mem = vg->vgmem;
 
 	/*
-	 * The pv hash memorises the pv section names -> pv
+	 * The pv hash memorizes the pv section names -> pv
 	 * structures.
 	 */
-	if (!(pv_hash = dm_hash_create(64))) {
+	if (!(pv_hash = dm_hash_create(59))) {
 		log_error("Couldn't create pv hash table.");
 		goto bad;
 	}
 
 	/*
-	 * The lv hash memorises the lv section names -> lv
+	 * The lv hash memorizes the lv section names -> lv
 	 * structures.
 	 */
-	if (!(lv_hash = dm_hash_create(1024))) {
+	if (!(lv_hash = dm_hash_create(1023))) {
 		log_error("Couldn't create lv hash table.");
 		goto bad;
 	}
@@ -1282,6 +1304,7 @@ static int _read_vgsummary(const struct format_type *fmt, const struct dm_config
 	const struct dm_config_node *vgn;
 	struct dm_pool *mem = fmt->cmd->mem;
 	const char *str;
+	struct id id;
 
 	if (!dm_config_get_str(cft->root, "creation_host", &str))
 		str = "";
@@ -1302,11 +1325,13 @@ static int _read_vgsummary(const struct format_type *fmt, const struct dm_config
 
 	vgn = vgn->child;
 
-	if (!_read_id(&vgsummary->vgid, vgn, "id")) {
+	if (!_read_id(&id, vgn, "id")) {
 		log_error("Couldn't read uuid for volume group %s.", vgsummary->vgname);
 		return 0;
 	}
 
+	memcpy(vgsummary->vgid, &id, ID_LEN);
+
 	if (!_read_flag_config(vgn, &vgsummary->vgstatus, VG_FLAGS)) {
 		log_error("Couldn't find status flags for volume group %s.",
 			  vgsummary->vgname);
@@ -1335,14 +1360,14 @@ static int _read_vgsummary(const struct format_type *fmt, const struct dm_config
 	return 1;
 }
 
-static struct text_vg_version_ops _vsn1_ops = {
+static const struct text_vg_version_ops _vsn1_ops = {
 	.check_version = _vsn1_check_version,
 	.read_vg = _read_vg,
 	.read_desc = _read_desc,
 	.read_vgsummary = _read_vgsummary
 };
 
-struct text_vg_version_ops *text_vg_vsn1_init(void)
+const struct text_vg_version_ops *text_vg_vsn1_init(void)
 {
 	return &_vsn1_ops;
 }
