@@ -31,7 +31,7 @@ int lv_is_origin(const struct logical_volume *lv)
 int lv_is_cow(const struct logical_volume *lv)
 {
 	/* Make sure a merging thin origin isn't confused as a cow LV */
-	return (!lv_is_thin_volume(lv) && !lv_is_origin(lv) && lv->snapshot) ? 1 : 0;
+	return (lv->snapshot && !lv_is_thin_volume(lv) && !lv_is_origin(lv)) ? 1 : 0;
 }
 
 struct logical_volume *find_cow(const struct logical_volume *snap)
@@ -111,13 +111,18 @@ int cow_has_min_chunks(const struct volume_group *vg, uint32_t cow_extents, uint
 
 int lv_is_cow_covering_origin(const struct logical_volume *lv)
 {
-	return lv_is_cow(lv) &&
-		(lv->size >= _cow_max_size(lv->vg->cmd, origin_from_cow(lv)->size,
-					   find_snapshot(lv)->chunk_size));
+	const struct logical_volume *origin;
+
+	return (lv_is_cow(lv) &&
+		(origin = origin_from_cow(lv)) &&
+		(lv->size >= _cow_max_size(lv->vg->cmd, origin->size,
+					   find_snapshot(lv)->chunk_size)));
 }
 
 int lv_is_visible(const struct logical_volume *lv)
 {
+	const struct logical_volume *origin;
+
 	if (lv_is_historical(lv))
 		return 1;
 
@@ -125,13 +130,16 @@ int lv_is_visible(const struct logical_volume *lv)
 		return 0;
 
 	if (lv_is_cow(lv)) {
-		if (lv_is_virtual_origin(origin_from_cow(lv)))
+		if (!(origin = origin_from_cow(lv)))
+			return_0;
+
+		if (lv_is_virtual_origin(origin))
 			return 1;
 
 		if (lv_is_merging_cow(lv))
 			return 0;
 
-		return lv_is_visible(origin_from_cow(lv));
+		return lv_is_visible(origin);
 	}
 
 	return lv->status & VISIBLE_LV ? 1 : 0;
@@ -160,6 +168,9 @@ struct logical_volume *origin_from_cow(const struct logical_volume *lv)
 {
 	if (lv->snapshot)
 		return lv->snapshot->origin;
+
+	log_debug(INTERNAL_ERROR "Cannot get origin from snapshot %s.",
+		  display_lvname(lv));
 	return NULL;
 }
 
@@ -286,8 +297,15 @@ int vg_add_snapshot(struct logical_volume *origin,
 
 int vg_remove_snapshot(struct logical_volume *cow)
 {
-	struct logical_volume *origin = origin_from_cow(cow);
-	int is_origin_active = lv_is_active(origin);
+	struct logical_volume *origin;
+	int is_origin_active;
+
+	if (!lv_is_cow(cow))
+		return_0;
+
+	origin = origin_from_cow(cow);
+
+	is_origin_active = lv_is_active(origin);
 
 	if (is_origin_active &&
 	    lv_is_virtual_origin(origin)) {
@@ -312,7 +330,7 @@ int vg_remove_snapshot(struct logical_volume *cow)
 		clear_snapshot_merge(origin);
 		/*
 		 * preload origin IFF "snapshot-merge" target is active
-		 * - IMPORTANT: avoids preload if inactivate merge is pending
+		 * - IMPORTANT: avoids preload if inactive merge is pending
 		 */
 	}
 
@@ -325,6 +343,16 @@ int vg_remove_snapshot(struct logical_volume *cow)
 	cow->snapshot = NULL;
 	lv_set_visible(cow);
 
+	/* When origin with all its snapshots is going to be remove
+	 * don't bother with individual manipulation with COWs
+	 * Note: removal proceeds only when origin is inactive */
+	if (is_origin_active && origin->to_remove) {
+		origin->vg->needs_write_and_commit = 1;
+		log_debug_metadata("Postponing write and commit for remove of snapshot %s.",
+				   display_lvname(cow));
+		return 1;
+	}
+
 	if (!vg_write(origin->vg))
 		return_0;
 
@@ -363,6 +391,14 @@ int validate_snapshot_origin(const struct logical_volume *origin_lv)
 {
 	const char *err = NULL; /* For error string */
 
+	if (lv_is_cache(origin_lv) || lv_is_writecache(origin_lv)) {
+		struct logical_volume *lv = seg_lv(first_seg(origin_lv), 0);
+		if (lv_is_raid(lv) && lv_raid_has_integrity(lv)) {
+			err = "raid with integrity";
+			goto out;
+		}
+	}
+
 	if (lv_is_cow(origin_lv))
 		err = "snapshots";
 	else if (lv_is_locked(origin_lv))
@@ -387,10 +423,9 @@ int validate_snapshot_origin(const struct logical_volume *origin_lv)
 		}
 	} else if (lv_is_raid_type(origin_lv) && !lv_is_raid(origin_lv)) {
 		err = "raid subvolumes";
-	} else if (lv_is_raid(origin_lv) && lv_raid_has_integrity((struct logical_volume *)origin_lv)) {
-		err = "raid with integrity";
 	}
 
+ out:
 	if (err) {
 		log_error("Snapshots of %s are not supported.", err);
 		return 0;
