@@ -87,7 +87,7 @@ mkdtemp() {
 	base_template=$(echo "$template" | sed 's/XX*$//')
 
 	# Calculate how many X's we've just removed.
-	nx=$(expr length "$template" - length "$base_template")
+	nx=$(( ${#template} - ${#base_template} ))
 
 	err=
 	i=1
@@ -138,25 +138,29 @@ STACKTRACE() {
 		# Check for all cores newer then TESTNAME file
 		# Assume users keep prefix 'core'
 		# TODO: possibly better integrate with coredumpctl & systemd
-		while IFS= read -r i; do
+		local cores=()
+		local IFS=$IFS_NL
+		cores=( $(find . "$(dirname "$(sysctl -n kernel.core_pattern)")" \
+			"/var/lib/systemd/coredump/" -name 'core*' -newer TESTNAME 2>/dev/null || true ) )
+
+		for i in "${cores[@]-}"; do
 			bin=$(gdb -batch -c "$i" 2>&1 | grep "generated by" | \
 			sed -e "s,.*generated by \`\([^ ']*\).*,\1,") || continue
-			{
-				echo bt full
-				echo l
-				echo quit
-			} > gdb_commands.txt || rm -f gdb_commands.txt
+			cat > gdb_commands.txt <<- EOF || rm -f gdb_commands.txt
+				bt full
+				l
+				quit
+			EOF
 
 			if test ! -s gdb_commands.txt ; then
-				echo "Out of disk space, can't check coredump $i generated by $bin."
-				continue
+				echo "Out of disk space, cannot check coredump \"$i\" generated by \"$bin\"."
+				break
 			fi
 
 			echo "## Checking coredump: $i generated by $bin."
 			gdb -batch -c "$i" -x gdb_commands.txt "$(which "$bin")" 2>/dev/null | \
 			sed -e "s,^,## GDB:	," || continue
-		done < <(find . "$(dirname "$(sysctl -n kernel.core_pattern)")" \
-			"/var/lib/systemd/coredump/" -name 'core*' -newer TESTNAME 2>/dev/null || true)
+		done
 	fi
 
 	test -f SKIP_THIS_TEST && exit 200
@@ -189,14 +193,19 @@ STACKTRACE() {
 			echo "<======== Tree ========>"
 			dmsetup ls --tree | sed -e "s,^,## DMTREE:   ,"
 			echo "<======== Recursive list of $DM_DEV_DIR ========>"
-			ls -Rl --hide=shm --hide=bus --hide=snd --hide=input --hide=dri \
-			       --hide=net --hide=hugepages --hide=mqueue --hide=pts \
-			       "$DM_DEV_DIR" | sed -e "s,^,## LSLR:	,"
+			ls -lR -I bsg -I bus -I char -Idma_heap -I dri \
+			   -I hugepages -I input -I mqueue \
+			   -I net -I pts -I shm -I snd \
+			   -I tty?* -I usb -I vfio -I vcs?* \
+			   -I virtio-ports \
+			   "$DM_DEV_DIR" | sed -e "s,^,## LS_LR:	,"
 			echo "<======== Udev DB content ========>"
 			for i in /sys/block/dm-* /sys/block/loop* ; do
 				udevadm info --query=all --path "$i" 2>/dev/null || true
 			done | sed -e "s,^,## UDEV:	,"
 		fi
+		echo "<======== Free space ========>"
+		df -h | sed -e "s,^,## DF_H:	,"
 		echo "<======== Script file \"$(< TESTNAME)\" ========>"
 		local script=$0
 		test -f "$script" || script="$TESTOLDPWD/$0"
@@ -205,7 +214,7 @@ STACKTRACE() {
 }
 
 init_udev_transaction() {
-	if test "$DM_UDEV_SYNCHRONISATION" = 1; then
+	if test "$DM_UDEV_SYNCHRONIZATION" = 1; then
 		local cookie
 		cookie=$(dmsetup udevcreatecookie)
 		# Cookie is not generated if udev is not running!
@@ -214,14 +223,14 @@ init_udev_transaction() {
 }
 
 finish_udev_transaction() {
-	if test "$DM_UDEV_SYNCHRONISATION" = 1 && test -n "${DM_UDEV_COOKIE-}" ; then
+	if test "$DM_UDEV_SYNCHRONIZATION" = 1 && test -n "${DM_UDEV_COOKIE-}" ; then
 		dmsetup udevreleasecookie || true
 		unset DM_UDEV_COOKIE
 	fi
 }
 
 teardown_udev_cookies() {
-	if test "$DM_UDEV_SYNCHRONISATION" = 1; then
+	if test "$DM_UDEV_SYNCHRONIZATION" = 1; then
 		# Delete any cookies created more than 10 minutes ago
 		# and not used in the last 10 minutes.
 		# Log only non-zero semaphores count
@@ -276,7 +285,7 @@ prepare_test_vars() {
 
 	if test -n "$LVM_TEST_DEVICE_LIST"; then
 		local count=0
-		while read path; do
+		while read -r path; do
 			count=$((  count + 1 ))
 			eval "dev$count=\"$path\""
 		done < "$LVM_TEST_DEVICE_LIST"
@@ -297,9 +306,7 @@ if test -z "${installed_testsuite+varset}"; then
     *)
 	PATH="$abs_top_builddir/test/lib:$abs_top_builddir/test/api:$PATH"
 	LVM_BINARY=$(which lvm)
-	LD_LIBRARY_PATH=$(find -L "$abs_top_builddir/libdm/" "$abs_top_builddir/tools/"\
-		"$abs_top_builddir/daemons/" \
-		-name "*.so" -printf "%h:")"$LD_LIBRARY_PATH"
+	LD_LIBRARY_PATH="$abs_top_builddir/daemons/dmeventd:$abs_top_builddir/tools:$abs_top_builddir/libdm:$LD_LIBRARY_PATH"
 	export PATH LD_LIBRARY_PATH LVM_BINARY ;;
     esac
 fi
