@@ -19,10 +19,11 @@
 struct cmd_context;
 
 struct cmdline_context {
-	struct opt_name *opt_names;
+	const struct opt_name *opt_names;
 	struct command *commands;
 	int num_commands;
-	struct command_name *command_names;
+	const struct command_name *command_names;
+	const struct command_name_args *command_names_args;
 	int num_command_names;
 };
 
