@@ -19,6 +19,7 @@
 #include "lib/format_text/layout.h"
 #include "lib/mm/xlate.h"
 #include "lib/misc/crc.h"
+#include "lib/device/device_id.h"
 
 #define ONE_MB_IN_BYTES 1048576
 
@@ -39,7 +40,7 @@ struct settings {
 	uint64_t device_size;      /* bytes */
 	uint64_t data_offset;      /* bytes, start of data (pe_start) */
 	uint32_t seqno;
-	struct id pvid;
+	struct id pv_id;
 
 	int mda_num;               /* 1 or 2 for first or second mda */
 	char *backup_file;
@@ -283,7 +284,7 @@ static int _text_buf_parsable(char *text_buf, uint64_t text_size)
 #define MAX_LINE_CHECK 128
 
 #define MAX_DESC 1024
-char desc_line[MAX_DESC];
+static char _desc_line[MAX_DESC];
 
 static void _copy_line(char *in, char *out, int *len, int linesize)
 {
@@ -342,14 +343,15 @@ static uint64_t mda2_size_from_offset(struct device *dev, uint64_t mda2_offset)
 }
 
 struct devicefile {
-	char path[PATH_MAX];
 	int fd;
+	char path[0];
 };
 
-static struct devicefile *get_devicefile(const char *path)
+static struct devicefile *get_devicefile(struct cmd_context *cmd, const char *path)
 {
 	struct stat sb;
 	struct devicefile *def;
+	size_t len;
 
 	if (stat(path, &sb))
 		return_NULL;
@@ -357,18 +359,14 @@ static struct devicefile *get_devicefile(const char *path)
 	if ((sb.st_mode & S_IFMT) != S_IFREG)
 		return_NULL;
 
-	if (!(def = malloc(sizeof(struct devicefile))))
+	len = strlen(path) + 1;
+	if (!(def = dm_pool_alloc(cmd->mem, sizeof(struct devicefile) + len)))
 		return_NULL;
 
-	if (dm_snprintf(def->path, PATH_MAX, "%s", path) < 0) {
-		free(def);
-		return_NULL;
-	}
+	memcpy(def->path, path, len);
 
-	if ((def->fd = open(path, O_RDONLY)) < 0) {
-		free(def);
+	if ((def->fd = open(path, O_RDONLY)) < 0)
 		return_NULL;
-	}
 
 	return def;
 }
@@ -381,14 +379,17 @@ static bool _read_bytes(struct device *dev, struct devicefile *def, uint64_t sta
 	if (dev)
 		return dev_read_bytes(dev, start, len, data);
 
+	if (!def)
+		return false;
+
 	off = lseek(def->fd, start, SEEK_SET);
-	if (off != start)
+	if (off != (off_t)start)
 		return false;
 
 	rv = read(def->fd, data, len);
 	if (rv < 0)
 		return false;
-	if (rv != len)
+	if ((size_t)rv != len)
 		return false;
 	return true;
 }
@@ -417,7 +418,7 @@ static int _dump_all_text(struct cmd_context *cmd, struct settings *set, const c
 	int multiple_vgs = 0;
 	int bad_end;
 	int vgnamelen;
-	int count;
+	unsigned count;
 	int len;
 
 	if (tofile) {
@@ -602,18 +603,18 @@ static int _dump_all_text(struct cmd_context *cmd, struct settings *set, const c
 		if (arg_is_set(cmd, verbose_ARG)) {
 			char *str1, *str2;
 			if ((str1 = strstr(text_buf, "description = "))) {
-				memset(desc_line, 0, sizeof(desc_line));
-				_copy_line(str1, desc_line, &len, sizeof(desc_line)-1);
-				if ((p = strchr(desc_line, '\n')))
+				memset(_desc_line, 0, sizeof(_desc_line));
+				_copy_line(str1, _desc_line, &len, sizeof(_desc_line)-1);
+				if ((p = strchr(_desc_line, '\n')))
 					*p = '\0';
-				log_print("%s", desc_line);
+				log_print("%s", _desc_line);
 			}
 			if (str1 && (str2 = strstr(str1, "creation_time = "))) {
-				memset(desc_line, 0, sizeof(desc_line));
-				_copy_line(str2, desc_line, &len, sizeof(desc_line)-1);
-				if ((p = strchr(desc_line, '\n')))
+				memset(_desc_line, 0, sizeof(_desc_line));
+				_copy_line(str2, _desc_line, &len, sizeof(_desc_line)-1);
+				if ((p = strchr(_desc_line, '\n')))
 					*p = '\0';
-				log_print("%s\n", desc_line);
+				log_print("%s\n", _desc_line);
 			}
 		}
 
@@ -781,7 +782,6 @@ static int _dump_raw_locn(struct device *dev, struct devicefile *def, int print_
 	uint64_t meta_offset, meta_size;
 	uint32_t meta_checksum;
 	uint32_t meta_flags;
-	int bad = 0;
 	int mn = mda_num; /* 1 or 2 */
 	int ri = rlocn_index; /* 0 or 1 */
 	int wrapped = 0;
@@ -836,8 +836,6 @@ static int _dump_raw_locn(struct device *dev, struct devicefile *def, int print_
 	if (!meta_offset)
 		return 1;
 
-	if (bad)
-		return 0;
 	return 1;
 }
 
@@ -851,7 +849,7 @@ static int _dump_meta_area(struct device *dev, struct devicefile *def, const cha
 	if (!tofile)
 		return_0;
 
-	if (!(meta_buf = zalloc(mda_size)))
+	if (!(meta_buf = zalloc(mda_size + 1)))
 		return_0;
 
 	if (!_read_bytes(dev, def, mda_offset, mda_size, meta_buf)) {
@@ -900,7 +898,7 @@ static int _dump_current_text(struct device *dev, struct devicefile *def,
 	int ri = rlocn_index; /* 0 or 1 */
 	int bad = 0;
 
-	if (!(meta_buf = zalloc(meta_size))) {
+	if (!(meta_buf = zalloc(meta_size + 1))) {
 		log_print("CHECK: mda_header_%d.raw_locn[%d] no mem for metadata text size %llu", mn, ri,
 			  (unsigned long long)meta_size);
 		return 0;
@@ -946,6 +944,7 @@ static int _dump_current_text(struct device *dev, struct devicefile *def,
 		}
 	}
 
+	meta_buf[meta_size] = 0;
 	crc = calc_crc(INITIAL_CRC, (uint8_t *)meta_buf, meta_size);
 	if (crc != meta_checksum) {
 		log_print("CHECK: metadata text at %llu crc does not match mda_header_%d.raw_locn[%d].checksum",
@@ -1016,7 +1015,7 @@ static int _dump_label_and_pv_header(struct cmd_context *cmd, uint64_t labelsect
 				     uint64_t *mda2_offset, uint64_t *mda2_size,
 				     int *mda_count_out)
 {
-	char buf[512];
+	char buf[512 + 1] = { 0 };
 	char str[256];
 	struct label_header *lh;
 	struct pv_header *pvh;
@@ -1026,7 +1025,6 @@ static int _dump_label_and_pv_header(struct cmd_context *cmd, uint64_t labelsect
 	uint64_t pvh_offset;    /* bytes */
 	uint64_t pvhe_offset;   /* bytes */
 	uint64_t dlocn_offset;  /* bytes */
-	uint64_t tmp;
 	int mda_count = 0;
 	int bad = 0;
 	int di;
@@ -1067,10 +1065,6 @@ static int _dump_label_and_pv_header(struct cmd_context *cmd, uint64_t labelsect
 	pvh = (struct pv_header *)(buf + 32);
 	pvh_offset = lh_offset + 32; /* from start of disk */
 
-	/* sanity check */
-	if ((void *)pvh != (void *)(buf + pvh_offset - lh_offset))
-		log_print("CHECK: problem with pv_header offset calculation");
-
 	if (print_fields) {
 		log_print("pv_header at %llu", (unsigned long long)pvh_offset);
 		log_print("pv_header.pv_uuid %s", _chars_to_str(pvh->pv_uuid, str, ID_LEN, 256, "pv_header.pv_uuid"));
@@ -1089,11 +1083,7 @@ static int _dump_label_and_pv_header(struct cmd_context *cmd, uint64_t labelsect
 	dlocn = pvh->disk_areas_xl;
 	dlocn_offset = pvh_offset + 40; /* from start of disk */
 
-	/* sanity check */
-	if ((void *)dlocn != (void *)(buf + dlocn_offset - lh_offset))
-		log_print("CHECK: problem with pv_header.disk_locn[%d] offset calculation", di);
-
-	while ((tmp = xlate64(dlocn->offset))) {
+	while (xlate64(dlocn->offset)) {
 		if (print_fields) {
 			log_print("pv_header.disk_locn[%d] at %llu # location of data area", di,
 				  (unsigned long long)dlocn_offset);
@@ -1126,7 +1116,7 @@ static int _dump_label_and_pv_header(struct cmd_context *cmd, uint64_t labelsect
 	if ((void *)dlocn != (void *)(buf + dlocn_offset - lh_offset))
 		log_print("CHECK: problem with pv_header.disk_locn[%d] offset calculation", di);
 
-	while ((tmp = xlate64(dlocn->offset))) {
+	while (xlate64(dlocn->offset)) {
 		if (print_fields) {
 			log_print("pv_header.disk_locn[%d] at %llu # location of metadata area", di,
 				  (unsigned long long)dlocn_offset);
@@ -1140,9 +1130,16 @@ static int _dump_label_and_pv_header(struct cmd_context *cmd, uint64_t labelsect
 			*mda1_offset = xlate64(dlocn->offset);
 			*mda1_size = xlate64(dlocn->size);
 
-			if (*mda1_offset != 4096) {
-				log_print("CHECK: pv_header.disk_locn[%d].offset expected 4096 # for first mda", di);
-				bad++;
+			/*
+			 * mda1 offset is page size from machine that created it,
+			 * warn if it's not one of the expected page sizes.
+			 */
+			if ((*mda1_offset != 4096) && 
+			    (*mda1_offset != 8192) &&
+			    (*mda1_offset != 16384) &&
+			    (*mda1_offset != 65536)) {
+				log_print("WARNING: pv_header.disk_locn[%d].offset %llu is unexpected # for first mda",
+					  di, (unsigned long long)*mda1_offset);
 			}
 		} else {
 			*mda2_offset = xlate64(dlocn->offset);
@@ -1205,7 +1202,7 @@ static int _dump_label_and_pv_header(struct cmd_context *cmd, uint64_t labelsect
 	dlocn = pvhe->bootloader_areas_xl;
 	dlocn_offset = pvhe_offset + 8;
 
-	while ((tmp = xlate64(dlocn->offset))) {
+	while (xlate64(dlocn->offset)) {
 		if (print_fields) {
 			log_print("pv_header_extension.disk_locn[%d] at %llu # bootloader area", di,
 				  (unsigned long long)dlocn_offset);
@@ -1258,7 +1255,7 @@ static int _dump_mda_header(struct cmd_context *cmd, struct settings *set,
 			    uint32_t *checksum0_ret,
 			    int *found_header)
 {
-	char buf[512];
+	char buf[512 + 1] = { 0 };
 	char str[256];
 	char *mda_buf;
 	struct mda_header *mh;
@@ -1267,7 +1264,7 @@ static int _dump_mda_header(struct cmd_context *cmd, struct settings *set,
 	uint64_t meta_offset = 0; /* bytes */
 	uint64_t meta_size = 0;   /* bytes */
 	uint32_t meta_checksum = 0;
-	int mda_num = (mda_offset == 4096) ? 1 : 2;
+	int mda_num = (mda_offset <= 65536) ? 1 : 2;
 	int bad = 0;
 
 	*checksum0_ret = 0; /* checksum from raw_locn[0] */
@@ -1357,7 +1354,7 @@ static int _dump_mda_header(struct cmd_context *cmd, struct settings *set,
 	 * looking at all copies of the metadata in the area
 	 */
 	if (print_metadata == PRINT_ALL) {
-		if (!(mda_buf = zalloc(mda_size)))
+		if (!(mda_buf = zalloc(mda_size + 1)))
 			goto_out;
 
 		if (!_read_bytes(dev, def, mda_offset, mda_size, mda_buf)) {
@@ -1435,8 +1432,13 @@ static int _dump_metadata(struct cmd_context *cmd, const char *dump, struct sett
 	int bad = 0;
 
 	if (arg_is_set(cmd, file_ARG)) {
+		struct stat sb;
 		if (!(tofile = arg_str_value(cmd, file_ARG, NULL)))
 			return 0;
+		if (!stat(tofile, &sb)) {
+			log_error("File already exists.");
+			return 0;
+		}
 	}
 
 	if (set->mda_num)
@@ -1715,7 +1717,7 @@ static int _dump_search(struct cmd_context *cmd, const char *dump, struct settin
 		log_print("Override defaults with --settings \"mda_offset=<bytes> mda_size=<bytes>\"");
 	} else {
 		log_error("No mda location.");
-		return_0;
+		return 0;
 	}
 
  search:
@@ -1734,7 +1736,7 @@ static int _dump_search(struct cmd_context *cmd, const char *dump, struct settin
 	log_print("Searching for metadata at offset %llu size %llu",
 		  (unsigned long long)mda_offset, (unsigned long long)mda_size);
 
-	if (!(buf = zalloc(mda_size)))
+	if (!(buf = zalloc(mda_size + 1)))
 		return_0;
 
 	if (!_read_bytes(dev, def, mda_offset, mda_size, buf)) {
@@ -1754,74 +1756,74 @@ static int _dump_search(struct cmd_context *cmd, const char *dump, struct settin
 
 static int _get_one_setting(struct cmd_context *cmd, struct settings *set, char *key, char *val)
 {
-	if (!strncmp(key, "metadata_offset", strlen("metadata_offset"))) {
+	if (!strncmp(key, "metadata_offset", sizeof("metadata_offset") - 1)) {
 		if (sscanf(val, "%llu", (unsigned long long *)&set->metadata_offset) != 1)
 			goto_bad;
 		set->metadata_offset_set = 1;
 		return 1;
 	}
 
-	if (!strncmp(key, "seqno", strlen("seqno"))) {
+	if (!strncmp(key, "seqno", sizeof("seqno") - 1)) {
 		if (sscanf(val, "%u", &set->seqno) != 1)
 			goto_bad;
 		set->seqno_set = 1;
 		return 1;
 	}
 
-	if (!strncmp(key, "backup_file", strlen("backup_file"))) {
-		if ((set->backup_file = strdup(val)))
+	if (!strncmp(key, "backup_file", sizeof("backup_file") - 1)) {
+		if ((set->backup_file = dm_pool_strdup(cmd->mem, val)))
 			return 1;
 		return 0;
 	}
 
-	if (!strncmp(key, "mda_offset", strlen("mda_offset"))) {
+	if (!strncmp(key, "mda_offset", sizeof("mda_offset") - 1)) {
 		if (sscanf(val, "%llu", (unsigned long long *)&set->mda_offset) != 1)
 			goto_bad;
 		set->mda_offset_set = 1;
 		return 1;
 	}
 
-	if (!strncmp(key, "mda_size", strlen("mda_size"))) {
+	if (!strncmp(key, "mda_size", sizeof("mda_size") - 1)) {
 		if (sscanf(val, "%llu", (unsigned long long *)&set->mda_size) != 1)
 			goto_bad;
 		set->mda_size_set = 1;
 		return 1;
 	}
 
-	if (!strncmp(key, "mda2_offset", strlen("mda2_offset"))) {
+	if (!strncmp(key, "mda2_offset", sizeof("mda2_offset") - 1)) {
 		if (sscanf(val, "%llu", (unsigned long long *)&set->mda2_offset) != 1)
 			goto_bad;
 		set->mda2_offset_set = 1;
 		return 1;
 	}
 
-	if (!strncmp(key, "mda2_size", strlen("mda2_size"))) {
+	if (!strncmp(key, "mda2_size", sizeof("mda2_size") - 1)) {
 		if (sscanf(val, "%llu", (unsigned long long *)&set->mda2_size) != 1)
 			goto_bad;
 		set->mda2_size_set = 1;
 		return 1;
 	}
 
-	if (!strncmp(key, "device_size", strlen("device_size"))) {
+	if (!strncmp(key, "device_size", sizeof("device_size") - 1)) {
 		if (sscanf(val, "%llu", (unsigned long long *)&set->device_size) != 1)
 			goto_bad;
 		set->device_size_set = 1;
 		return 1;
 	}
 
-	if (!strncmp(key, "data_offset", strlen("data_offset"))) {
+	if (!strncmp(key, "data_offset", sizeof("data_offset") - 1)) {
 		if (sscanf(val, "%llu", (unsigned long long *)&set->data_offset) != 1)
 			goto_bad;
 		set->data_offset_set = 1;
 		return 1;
 	}
 
-	if (!strncmp(key, "pv_uuid", strlen("pv_uuid"))) {
+	if (!strncmp(key, "pv_uuid", sizeof("pv_uuid") - 1)) {
 		if (strchr(val, '-') && (strlen(val) == 32)) {
-			memcpy(&set->pvid, val, 32);
+			memcpy(&set->pv_id, val, 32);
 			set->pvid_set = 1;
 			return 1;
-		} else if (id_read_format_try(&set->pvid, val)) {
+		} else if (id_read_format_try(&set->pv_id, val)) {
 			set->pvid_set = 1;
 			return 1;
 		} else {
@@ -1830,7 +1832,7 @@ static int _get_one_setting(struct cmd_context *cmd, struct settings *set, char
 		}
 	}
 
-	if (!strncmp(key, "mda_num", strlen("mda_num"))) {
+	if (!strncmp(key, "mda_num", sizeof("mda_num") - 1)) {
 		if (sscanf(val, "%u", (int *)&set->mda_num) != 1)
 			goto_bad;
 		return 1;
@@ -1846,8 +1848,8 @@ static int _get_settings(struct cmd_context *cmd, struct settings *set)
 	const char *str;
 	char key[64];
 	char val[64];
-	int num;
-	int pos;
+	unsigned num;
+	unsigned pos;
 
 	/*
 	 * "grouped" means that multiple --settings options can be used.
@@ -1969,19 +1971,15 @@ static int _get_pv_info_from_metadata(struct cmd_context *cmd, struct settings *
 				     uint64_t *device_size_sectors,
 				     uint64_t *pe_start_sectors)
 {
-	int8_t pvid_cur[ID_LEN+1];  /* found in existing pv_header */
-	int8_t pvid_set[ID_LEN+1];  /* set by user in --settings */
-	int8_t pvid_use[ID_LEN+1];  /* the pvid chosen to use */
-	int pvid_cur_valid = 0;     /* pvid_cur is valid */
-	int pvid_use_valid = 0;     /* pvid_use is valid */
+	char pvid_cur[ID_LEN + 1] = { 0 };  /* found in existing pv_header */
+	char pvid_set[ID_LEN + 1] = { 0 };  /* set by user in --settings */
+	char pvid_use[ID_LEN + 1] = { 0 };  /* the pvid chosen to use */
+	int pvid_cur_valid = 0;             /* pvid_cur is valid */
+	int pvid_use_valid = 0;             /* pvid_use is valid */
 	struct dm_config_tree *cft = NULL;
 	struct volume_group *vg = NULL;
 	struct pv_list *pvl;
 
-	memset(pvid_cur, 0, sizeof(pvid_cur));
-	memset(pvid_set, 0, sizeof(pvid_set));
-	memset(pvid_use, 0, sizeof(pvid_use));
-
 	/*
 	 * Check if there's a valid existing PV UUID at the expected location.
 	 */
@@ -1994,14 +1992,14 @@ static int _get_pv_info_from_metadata(struct cmd_context *cmd, struct settings *
 	}
 
 	if (set->pvid_set) {
-		memcpy(&pvid_set, &set->pvid, ID_LEN);
+		memcpy(&pvid_set, &set->pv_id, ID_LEN);
 		memcpy(&pvid_use, &pvid_set, ID_LEN);
 		pvid_use_valid = 1;
 	}
 
 	if (pvid_cur_valid && set->pvid_set && memcmp(&pvid_cur, &pvid_set, ID_LEN)) {
 		log_warn("WARNING: existing PV UUID %s does not match pv_uuid setting %s.",
-			 (char *)&pvid_cur, (char *)&pvid_set);
+			 pvid_cur, pvid_set);
 
 		memcpy(&pvid_use, &pvid_set, ID_LEN);
 		pvid_use_valid = 1;
@@ -2039,7 +2037,7 @@ static int _get_pv_info_from_metadata(struct cmd_context *cmd, struct settings *
 		}
 	} else {
 		dm_list_iterate_items(pvl, &vg->pvs) {
-			if (id_equal(&pvl->pv->id, (struct id *)&pvid_use))
+			if (!memcmp(&pvl->pv->id.uuid, &pvid_use, ID_LEN))
 				goto copy_pv;
 		}
 	}
@@ -2053,9 +2051,9 @@ static int _get_pv_info_from_metadata(struct cmd_context *cmd, struct settings *
 	 * . the metadata has no PV with a device name hint matching this device
 	 */
 	if (set->pvid_set)
-		log_error("PV UUID %s not found in metadata file.", (char *)&pvid_set);
+		log_error("PV UUID %s not found in metadata file.", pvid_set);
 	else if (pvid_cur_valid)
-		log_error("PV UUID %s in existing pv_header not found in metadata file.", (char *)&pvid_cur);
+		log_error("PV UUID %s in existing pv_header not found in metadata file.", pvid_cur);
 	else if (!pvid_use_valid)
 		log_error("PV name %s not found in metadata file.", dev_name(dev));
 
@@ -2110,7 +2108,7 @@ static int _check_for_mda2(struct cmd_context *cmd, struct device *dev,
 	char buf2[256];
 	char *buf;
 	uint64_t mda_offset, mda_size, extra_bytes; /* bytes */
-	int i, found = 0;
+	unsigned i, found = 0;
 
 	if (device_size < (2 * ONE_MB_IN_BYTES))
 		return_0;
@@ -2214,7 +2212,7 @@ static int _repair_pv_header(struct cmd_context *cmd, const char *repair,
 			     uint64_t labelsector, struct device *dev)
 {
 	char head_buf[512];
-	int8_t pvid[ID_LEN+1];
+	char pvid[ID_LEN + 1] __attribute__((aligned(8))) = { 0 };
 	struct device *dev_with_pvid = NULL;
 	struct label_header *lh;
 	struct pv_header *pvh;
@@ -2232,8 +2230,6 @@ static int _repair_pv_header(struct cmd_context *cmd, const char *repair,
 	int found_label = 0;
 	int di;
 
-	memset(&pvid, 0, ID_LEN+1);
-
 	lh_offset = labelsector * 512; /* from start of disk */
 
 	if (!dev_get_size(dev, &get_size_sectors))
@@ -2344,7 +2340,7 @@ static int _repair_pv_header(struct cmd_context *cmd, const char *repair,
 	if (set->device_size_set && set->pvid_set && set->data_offset_set && !mf->filename) {
 		device_size = set->device_size;
 		pe_start_sectors = set->data_offset >> SECTOR_SHIFT;
-		memcpy(&pvid, &set->pvid, ID_LEN);
+		memcpy(pvid, &set->pv_id, ID_LEN);
 
 		if (get_size && (get_size != device_size)) {
 			log_warn("WARNING: device_size setting %llu bytes does not match device size %llu bytes.",
@@ -2376,7 +2372,7 @@ static int _repair_pv_header(struct cmd_context *cmd, const char *repair,
 	 * dev_size and pe_start from the metadata to use in the pv_header.
 	 */
 	if (!_get_pv_info_from_metadata(cmd, set, dev, pvh, found_label,
-					mf->text_buf, mf->text_size, (char *)&pvid,
+					mf->text_buf, mf->text_size, pvid,
 					&device_size_sectors, &pe_start_sectors))
 		goto fail;
 
@@ -2390,13 +2386,13 @@ static int _repair_pv_header(struct cmd_context *cmd, const char *repair,
 	 * Read all devs to verify the pvid that will be written does not exist
 	 * on another device.
 	 */
-	if (!label_scan_for_pvid(cmd, (char *)&pvid, &dev_with_pvid)) {
+	if (!label_scan_for_pvid(cmd, pvid, &dev_with_pvid)) {
 		log_error("Failed to scan devices to check PV UUID.");
 		goto fail;
 	}
 
 	if (dev_with_pvid && (dev_with_pvid != dev)) {
-		log_error("Cannot use PV UUID %s which exists on %s", (char *)&pvid, dev_name(dev_with_pvid));
+		log_error("Cannot use PV UUID %s which exists on %s", pvid, dev_name(dev_with_pvid));
 		goto fail;
 	}
 
@@ -2464,7 +2460,7 @@ static int _repair_pv_header(struct cmd_context *cmd, const char *repair,
 	 */
 
 	log_print("Writing label_header.crc 0x%08x pv_header uuid %s device_size %llu",
-		  head_crc, (char *)&pvid, (unsigned long long)device_size);
+		  head_crc, pvid, (unsigned long long)device_size);
 
 	log_print("Writing data_offset %llu mda1_offset %llu mda1_size %llu mda2_offset %llu mda2_size %llu",
 		  (unsigned long long)data_offset,
@@ -2826,7 +2822,7 @@ static int _dump_backup_to_raw(struct cmd_context *cmd, struct settings *set)
 		goto fail_close;
 
 	rv = read(fd, back_buf, back_size);
-	if (rv != back_size) {
+	if (rv != (int)back_size) {
 		log_error("Cannot read file: %s", input);
 		free(back_buf);
 		goto fail_close;
@@ -2959,17 +2955,16 @@ static int _read_metadata_file(struct cmd_context *cmd, struct metadata_file *mf
 		goto out;
 	}
 
-	if (!(text_buf = zalloc(text_size + 1)))
+	if (!(text_buf = malloc(text_size + 1)))
 		goto_out;
 
 	rv = read(fd, text_buf, text_size);
-	if (rv != text_size) {
+	if (rv != (int)text_size) {
 		log_error("Cannot read file: %s", mf->filename);
 		free(text_buf);
 		goto out;
 	}
-
-	text_size += 1; /* null terminating byte */
+	text_buf[text_size++] = 0; /* null terminating byte */
 
 	if (close(fd))
 		stack;
@@ -3004,22 +2999,18 @@ out:
 	return 0;
 }
 
-int pvck(struct cmd_context *cmd, int argc, char **argv)
+static int _pvck_mf(struct metadata_file *mf, struct cmd_context *cmd, int argc, char **argv)
 {
-	struct settings set;
-	struct metadata_file mf;
+	struct settings set = { 0 };
 	struct device *dev = NULL;
 	struct devicefile *def = NULL;
 	const char *dump, *repair;
-	const char *pv_name;
+	const char *pv_name = "";
 	uint64_t labelsector = 1;
 	int bad = 0;
 	int ret = 0;
 	int i;
 
-	memset(&set, 0, sizeof(set));
-	memset(&mf, 0, sizeof(mf));
-
 	/*
 	 * By default LVM skips the first sector (sector 0), and writes
 	 * the label_header in the second sector (sector 1).
@@ -3031,22 +3022,40 @@ int pvck(struct cmd_context *cmd, int argc, char **argv)
 	if (arg_is_set(cmd, repairtype_ARG) || arg_is_set(cmd, repair_ARG)) {
 		pv_name = argv[0];
 
+		if (!lock_global(cmd, "ex"))
+			return ECMD_FAILED;
+
 		clear_hint_file(cmd);
 
+		if (!setup_device(cmd, pv_name)) {
+			log_error("Failed to set up device %s.", pv_name);
+			return ECMD_FAILED;
+		}
+
 		if (!(dev = dev_cache_get(cmd, pv_name, NULL))) {
 			log_error("Cannot use %s: %s.", pv_name, devname_error_reason(pv_name));
 			return ECMD_FAILED;
 		}
 	}
 
-	if (arg_is_set(cmd, dump_ARG)) {
+	if ((dump = arg_str_value(cmd, dump_ARG, NULL))) {
+		struct stat sb;
+
 		pv_name = argv[0];
 
-		dev = dev_cache_get(cmd, pv_name, NULL);
-
-		if (!dev)
-			def = get_devicefile(pv_name);
-
+		if (stat(pv_name, &sb) < 0) {
+			log_error("Cannot access %s.", pv_name);
+			return ECMD_FAILED;
+		}
+		if (S_ISREG(sb.st_mode))
+			def = get_devicefile(cmd, pv_name);
+		else if (S_ISBLK(sb.st_mode)) {
+			if (!setup_device(cmd, pv_name)) {
+				log_error("Failed to set up device %s.", pv_name);
+				return ECMD_FAILED;
+			}
+			dev = dev_cache_get(cmd, pv_name, NULL);
+		}
 		if (!dev && !def) {
 			log_error("Cannot use %s: %s.", pv_name, devname_error_reason(pv_name));
 			return ECMD_FAILED;
@@ -3054,14 +3063,14 @@ int pvck(struct cmd_context *cmd, int argc, char **argv)
 	}
 
 	if (!_get_settings(cmd, &set))
-		return ECMD_FAILED;
+		return_ECMD_FAILED;
 
 	if (arg_is_set(cmd, file_ARG) && (arg_is_set(cmd, repairtype_ARG) || arg_is_set(cmd, repair_ARG))) {
-		if (!(mf.filename = arg_str_value(cmd, file_ARG, NULL)))
-			return ECMD_FAILED;
+		if (!(mf->filename = arg_str_value(cmd, file_ARG, NULL)))
+			return_ECMD_FAILED;
 
-		if (!_read_metadata_file(cmd, &mf))
-			return ECMD_FAILED;
+		if (!_read_metadata_file(cmd, mf))
+			return_ECMD_FAILED;
 	}
 
 	label_scan_setup_bcache();
@@ -3069,12 +3078,25 @@ int pvck(struct cmd_context *cmd, int argc, char **argv)
 	if (dev) {
 		char buf[4096];
 
+		/*
+		 * Check nodata filters including device_id filter,
+		 * then clear the result so the full filter can be
+		 * checked after reading the dev.
+		 */
+		cmd->filter_nodata_only = 1;
+		if (!cmd->filter->passes_filter(cmd, cmd->filter, dev, NULL)) {
+			log_error("Cannot use %s: %s.", pv_name, dev_filtered_reason(dev));
+			return ECMD_FAILED;
+		}
+		cmd->filter_nodata_only = 0;
+		cmd->filter->wipe(cmd, cmd->filter, dev, NULL);
+
 		/*
 		 * This buf is not used, but bcache data is used for subsequent
 		 * reads in the filters and by _read_bytes for other disk structs.
 		 */
 		if (!dev_read_bytes(dev, 0, 4096, buf)) {
-	        	log_error("Failed to read the first 4096 bytes of device %s.", dev_name(dev));
+			log_error("Failed to read the first 4096 bytes of device %s.", dev_name(dev));
 			return ECMD_FAILED;
 		}
 
@@ -3084,7 +3106,7 @@ int pvck(struct cmd_context *cmd, int argc, char **argv)
 		}
 	}
 
-	if ((dump = arg_str_value(cmd, dump_ARG, NULL))) {
+	if (dump) {
 		cmd->use_hints = 0;
 
 		if (!strcmp(dump, "metadata"))
@@ -3109,7 +3131,7 @@ int pvck(struct cmd_context *cmd, int argc, char **argv)
 			log_error("Unknown dump value.");
 
 		if (!ret)
-			return ECMD_FAILED;
+			return_ECMD_FAILED;
 		return ECMD_PROCESSED;
 	}
 
@@ -3120,15 +3142,15 @@ int pvck(struct cmd_context *cmd, int argc, char **argv)
 			ret = _repair_label_header(cmd, repair, &set, labelsector, dev);
 
 		else if (!strcmp(repair, "pv_header"))
-			ret = _repair_pv_header(cmd, repair, &set, &mf, labelsector, dev);
+			ret = _repair_pv_header(cmd, repair, &set, mf, labelsector, dev);
 
 		else if (!strcmp(repair, "metadata"))
-			ret = _repair_metadata(cmd, repair, &set, &mf, labelsector, dev);
+			ret = _repair_metadata(cmd, repair, &set, mf, labelsector, dev);
 		else
 			log_error("Unknown repair value.");
 
 		if (!ret)
-			return ECMD_FAILED;
+			return_ECMD_FAILED;
 		return ECMD_PROCESSED;
 	}
 
@@ -3137,11 +3159,11 @@ int pvck(struct cmd_context *cmd, int argc, char **argv)
 
 		/* repair is a combination of repairtype pv_header+metadata */
 
-		if (!_repair_pv_header(cmd, "pv_header", &set, &mf, labelsector, dev))
-			return ECMD_FAILED;
+		if (!_repair_pv_header(cmd, "pv_header", &set, mf, labelsector, dev))
+			return_ECMD_FAILED;
 
-		if (!_repair_metadata(cmd, "metadata", &set, &mf, labelsector, dev))
-			return ECMD_FAILED;
+		if (!_repair_metadata(cmd, "metadata", &set, mf, labelsector, dev))
+			return_ECMD_FAILED;
 
 		return ECMD_PROCESSED;
 	}
@@ -3151,6 +3173,12 @@ int pvck(struct cmd_context *cmd, int argc, char **argv)
 	 * but this is here to preserve the historical output.
 	 */
 
+	if (argc == 1) {
+		if (!setup_device(cmd, argv[0]))
+			return_ECMD_FAILED;
+	} else if (!setup_devices(cmd))
+		return_ECMD_FAILED;
+
 	for (i = 0; i < argc; i++) {
 		pv_name = argv[i];
 
@@ -3164,6 +3192,15 @@ int pvck(struct cmd_context *cmd, int argc, char **argv)
 	}
 
 	if (bad)
-		return ECMD_FAILED;
+		return_ECMD_FAILED;
 	return ECMD_PROCESSED;
 }
+
+int pvck(struct cmd_context *cmd, int argc, char **argv)
+{
+	struct metadata_file mf = { 0 };
+	int ret = _pvck_mf(&mf, cmd, argc, argv);
+
+	free(mf.text_buf);
+	return ret;
+}
